#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2020- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

from __future__ import print_function
import ibm_watson_machine_learning._wrappers.requests as requests

from ibm_watson_machine_learning.messages.messages import Messages
from ibm_watson_machine_learning.metanames import ConnectionMetaNames
from ibm_watson_machine_learning.wml_resource import WMLResource
from ibm_watson_machine_learning.wml_client_error import WMLClientError
import os
import json
from urllib.parse import unquote, quote

_DEFAULT_LIST_LENGTH = 0


class Connections(WMLResource):
    """Store and manage Connections."""

    ConfigurationMetaNames = ConnectionMetaNames()
    """MetaNames for Connection creation."""

    def __init__(self, client):
        WMLResource.__init__(self, __name__, client)
        self._ICP = client.ICP

    def _get_required_element_from_response(self, response_data):

        WMLResource._validate_type(response_data, u'connection_response', dict)
        try:
            new_el = {'metadata': {'id': response_data['metadata']['asset_id'],
                                   'asset_type': response_data['metadata']['asset_type'],
                                   'create_time': response_data['metadata']['create_time']
                                   if 'create_time' in response_data['metadata']
                                   else response_data['metadata']['created_at'],
                                   'last_access_time': response_data['metadata']['usage'].get('last_access_time')
                                   },
                      'entity': {
                          'datasource_type': response_data['entity']['datasource_type']
                          if 'datasource_type' in response_data['entity']
                          else response_data['entity']['connection']['datasource_type'],
                          'name': response_data['entity']['name']
                          if 'name' in response_data['entity']
                          else response_data['metadata']['name']
                      }
            }

            for el in ['description', 'origin_country', 'owner_id', 'properties']:
                if el in response_data['entity']:
                    new_el['entity'][el] = response_data['entity'].get(el)

            if self._client.default_space_id is not None:
                new_el['metadata']['space_id'] = response_data['metadata']['space_id']

            elif self._client.default_project_id is not None:
                new_el['metadata']['project_id'] = response_data['metadata']['project_id']

                if not self._client.WSD:
                    if 'href' in response_data['metadata']:
                        if self._client.CLOUD_PLATFORM_SPACES or self._client.ICP_PLATFORM_SPACES:
                            href_without_host = response_data['href'].split('.com')[-1]
                            new_el[u'metadata'].update({'href': href_without_host})
                        else:
                            new_el['metadata'].update({'href': response_data['href']})

            return new_el
        except Exception as e:
            raise e
            #raise WMLClientError(Messages.get_message(response_data, message_id="failed_to_read_response_from_down_stream_service"))

    def get_details(self, connection_id=None):
        """Get connection details for the given unique Connection id.
        If no connection_id is passed, details for all connections will be returned.

        :param connection_id: Unique id of Connection
        :type connection_id: str

        :return: metadata of the stored Connection
        :rtype: dict

        **Example**

        .. code-block:: python

            connection_details = client.connections.get_details(connection_id)
            connection_details = client.connections.get_details()

        """
        self._client._check_if_either_is_set()
        Connections._validate_type(connection_id, u'connection_id', str, False)

        if self._client.WSD:
            header_param = self._client._get_headers(wsdconnection_api_req=True)
        else:
            header_param = self._client._get_headers()

        if self._client._iam_id:
            header_param['IBM-WDP-Impersonate'] = str({'iam_id': str(self._client._iam_id)})

        if connection_id:
            with self.requests_retry_session() as sess:
                response = sess.get(
                    self._client.service_instance._href_definitions.get_connection_by_id_href(connection_id),
                    params=self._client._params(),
                    headers=header_param)

                return self._get_required_element_from_response(
                    self._handle_response(200, u'get connection details', response))
        else:
            with self.requests_retry_session() as sess:
                response = sess.post(
                    self._client.service_instance._href_definitions.get_asset_search_href('connection'),
                    json={"query": "*:*",
                          "include": "entity"
                          },
                    params=self._client._params(),
                    headers=header_param)

                return {'resources': [self._get_required_element_from_response(r) for r
                                      in self._handle_response(200, u'get connection details', response)['results']]}

    def create(self, meta_props):
        """Create a connection. Input to PROPERTIES field examples:

        1. MySQL

            .. code-block:: python

                client.connections.ConfigurationMetaNames.PROPERTIES: {
                    "database": "database",
                    "password": "password",
                    "port": "3306",
                    "host": "host url",
                    "ssl": "false",
                    "username": "username"
                }

        2. Google Big query

            a. Method1: Use service account json. The service account json generated can be provided as
                input as-is. Provide actual values in json. Example is only indicative to show
                the fields. Refer to Google big query documents how to generate the service account json.

                .. code-block:: python

                    client.connections.ConfigurationMetaNames.PROPERTIES: {
                        "type": "service_account",
                        "project_id": "project_id",
                        "private_key_id": "private_key_id",
                        "private_key": "private key contents",
                        "client_email": "client_email",
                        "client_id": "client_id",
                        "auth_uri": "https://accounts.google.com/o/oauth2/auth",
                        "token_uri": "https://oauth2.googleapis.com/token",
                        "auth_provider_x509_cert_url": "https://www.googleapis.com/oauth2/v1/certs",
                        "client_x509_cert_url": "client_x509_cert_url"
                    }

            b. Method2: Using OAuth Method. Refer to Google big query documents how to generate OAuth token.

                .. code-block:: python

                    client.connections.ConfigurationMetaNames.PROPERTIES: {
                        "access_token": "access token generated for big query",
                        "refresh_token": "refresh token",
                        "project_id": "project_id",
                        "client_secret": "This is your gmail account password",
                        "client_id": "client_id"
                    }

        3. MS SQL

            .. code-block:: python

                client.connections.ConfigurationMetaNames.PROPERTIES: {
                    "database": "database",
                    "password": "password",
                    "port": "1433",
                    "host": "host",
                    "username": "username"
                }

        4. Tera data

            .. code-block:: python

                client.connections.ConfigurationMetaNames.PROPERTIES: {
                    "database": "database",
                    "password": "password",
                    "port": "1433",
                    "host": "host",
                    "username": "username"
                }

        :param meta_props:  metadata of the connection configuration. To see available meta names use:

            .. code-block:: python

                client.connections.ConfigurationMetaNames.get()

        :type meta_props: dict

        :return: metadata of the stored connection
        :rtype: dict

        **Example**

        .. code-block:: python

            sqlserver_data_source_type_id = client.connections.get_datasource_type_uid_by_name('sqlserver')
            connections_details = client.connections.create({
                client.connections.ConfigurationMetaNames.NAME: "sqlserver connection",
                client.connections.ConfigurationMetaNames.DESCRIPTION: "connection description",
                client.connections.ConfigurationMetaNames.DATASOURCE_TYPE: sqlserver_data_source_type_id,
                client.connections.ConfigurationMetaNames.PROPERTIES: { "database": "database",
                                                                        "password": "password",
                                                                        "port": "1433",
                                                                        "host": "host",
                                                                        "username": "username"}
            })

        """
        WMLResource._chk_and_block_create_update_for_python36(self)

        connection_meta = self.ConfigurationMetaNames._generate_resource_metadata(
            meta_props,
            with_validation=True,
            client=self._client
        )

        big_query_data_source_type_id = self.get_datasource_type_uid_by_name('bigquery')

        # Either service acct json credentials can be given or oauth json can be given
        # If service acct json, then we need to create a newline json with "credentials" key
        if connection_meta[u'datasource_type'] == big_query_data_source_type_id:
            if 'private_key' in connection_meta[u'properties']:
                result = json.dumps(connection_meta[u'properties'], separators=(',\n', ':'))
                newmap = {"credentials": result}
                connection_meta[u'properties'] = newmap

        connection_meta.update({'origin_country': 'US'})
        #Step1  : Create an asset
        print(Messages.get_message(message_id="creating_connections"))

        if self._client.WSD:
            header_param = self._client._get_headers(wsdconnection_api_req=True)
        else:
            header_param = self._client._get_headers()

        creation_response = requests.post(
            self._client.service_instance._href_definitions.get_connections_href(),
            headers=header_param,
            json=connection_meta,
            params=self._client._params()
        )
        connection_details = self._handle_response(201, u'creating new connection', creation_response)
        if creation_response.status_code == 201:
            connection_id = connection_details["metadata"]["asset_id"]
            print(Messages.get_message(message_id="success"))
            return self._get_required_element_from_response(connection_details)
        else:
            raise WMLClientError(Messages.get_message(message_id="failed_while_creating_connections"))

    def delete(self, connection_id):
        """Delete a stored Connection.

        :param connection_id: Unique id of the connection to be deleted.
        :type connection_id: str

        :return: status ("SUCCESS" or "FAILED")
        :rtype: str

        **Example**

        .. code-block:: python

            client.connections.delete(connection_id)

        """
        ##For CP4D, check if either spce or project ID is set
        self._client._check_if_either_is_set()

        Connections._validate_type(connection_id, u'connection_id', str, True)
        if self._client.WSD:
            header_param = self._client._get_headers(wsdconnection_api_req=True)
        else:
            header_param = self._client._get_headers()

        connection_endpoint = self._client.service_instance._href_definitions.get_connection_by_id_href(connection_id)
        response_delete = requests.delete(connection_endpoint, params=self._client._params(),
                                          headers=header_param)

        return self._handle_response(204, u'connection deletion', response_delete, False)

    @staticmethod
    def get_uid(connection_details):
        """Get Unique Id of stored connection.

        :param connection_details: metadata of the stored connection
        :type connection_details: dict

        :return: Unique Id of stored connection
        :rtype: str

        **Example**

        .. code-block:: python

            connection_uid = client.connection.get_uid(connection_details)

        """
        Connections._validate_type(connection_details, u'connection_details', object, True)

        return WMLResource._get_required_element_from_dict(connection_details, u'connection_details',
                                                           [u'metadata', u'id'])

    def list_datasource_types(self):
        """Print stored datasource types assets in a table format.

        **Example**

        .. code-block:: python

            client.connections.list_datasource_types()

        """

        if self._client.WSD:
            header_param = self._client._get_headers(wsdconnection_api_req=True)
        else:
            header_param = self._client._get_headers()

        response = requests.get(self._client.service_instance._href_definitions.get_connection_data_types_href(),
                                headers=header_param)

        datasource_details = self._handle_response(200, u'list datasource types', response)['resources']
        space_values = [
            (m[u'entity'][u'name'], m[u'metadata'][u'asset_id'], m[u'entity'][u'type'], m['entity']['status']) for
            m in datasource_details]

        self._list(space_values, [u'NAME', u'DATASOURCE_ID', u'TYPE', u'STATUS'], None, None)

    def list(self):
        """Print all stored connections in a table format.

        **Example**

        .. code-block:: python

            client.connections.list()

        """
        datasource_details = self.get_details()
        space_values = [
            (m[u'entity'][u'name'], m[u'metadata'][u'id'], m['metadata']['create_time'],
             m['entity']['datasource_type']) for
            m in datasource_details['resources']]

        self._list(space_values, [u'NAME', u'ID', u'CREATED', 'DATASOURCE_TYPE_ID'], None, None)

    def list_uploaded_db_drivers(self):
        """Print uploaded db driver jars in table a format. Supported for IBM Cloud Pak for Data only.

        **Example**

        .. code-block:: python

            client.connections.list_uploaded_db_drivers()

        """
        if not self._client.ICP:
            raise WMLClientError('Not supported on this environment.')

        try:
            if not self.get_uploaded_db_drivers():
                raise Exception('List empty for new api')
            self._list_uploaded_db_drivers_new_api()
        except:
            response = requests.get(
                self._client.service_instance._href_definitions.get_wsd_model_attachment_href() + 'dbdrivers',
                headers=self._client._get_headers(no_content_type=True),
                params=self._client._params()
            )
            jars = [[el['path'].split('/')[-1]] for el in response.json()['resources']]

            self._list(jars, [u'NAME'], None, None)

    def get_uploaded_db_drivers(self): # TODO not yet documented
        # """
        #    Get uploaded db driver jar names and paths.
        #    Supported for IBM Cloud Pak for Data, version 4.6.1 and above.
        #
        #    **Output**
        #
        #    .. important::
        #         Returns dictionary containing name and path for connection files.\n
        #         **return type**: Dict[Str, Str]\n
        #
        #    **Example**
        #
        #     >>> result = client.connections.get_uploaded_db_drivers()
        #
        # """
        if not self._client.ICP:
            raise WMLClientError('Not supported on this environment.')

        response = requests.get(
            self._client.service_instance._href_definitions.get_connections_files_href(),
            headers=self._client._get_headers(no_content_type=True)
        )
        result = self._handle_response(200, u'get uploaded db drivers', response)['resources']
        return dict([(el['fileName'], el['url']) for el in result])

    def _list_uploaded_db_drivers_new_api(self):
        """List uploaded db driver jars. Supported for IBM Cloud Pak for Data only.

        .. important::
            This method only prints the uploaded db driver jar names.

        **Example**

        .. code-clock:: python

            client.connections._list_uploaded_db_drivers_new_api()

        """
        jars = [[name] for name in self.get_uploaded_db_drivers()]
        self._list(jars, [u'NAME'], None, None)

    def get_datasource_type_uid_by_name(self, name):
        """Get stored datasource types id for the given datasource type name.

        :param name:  name of datasource type
        :type name: int

        :return: datasource Unique Id
        :rtype: str

        **Example**

        .. code-block:: python

            client.connections.get_datasource_type_uid_by_name('cloudobjectstorage')

        """
        Connections._validate_type(name, u'name', str, True)

        if self._client.WSD:
            header_param = self._client._get_headers(wsdconnection_api_req=True)
        else:
            header_param = self._client._get_headers()

        with self.requests_retry_session() as sess:
            response = sess.get(self._client.service_instance._href_definitions.get_connection_data_types_href(),
                                headers=header_param)

            datasource_id = 'None'
            datasource_details = self._handle_response(200, u'list datasource types', response)['resources']

        for i, ds_resource in enumerate(datasource_details):
            if ds_resource['entity']['name'] == name:
                datasource_id = ds_resource['metadata']['asset_id']
        return datasource_id

    def upload_db_driver(self, path: str):
        """Upload db driver jar. Supported for IBM Cloud Pak for Data only, version 4.0.4 and above.

        :param path:  path to db driver jar
        :type path: str

        **Example**

        .. code-block:: python

            client.connections.upload_db_driver('example/path/db2jcc4.jar')

        """
        if not self._client.ICP:
            raise WMLClientError('Not supported on this environment.')

        try:
            self._upload_db_driver_new_api(path)
        except:
            driver_file_name = path.split('/')[-1]

            with open(path, 'rb') as fdata:
                content_upload_url = self._client.service_instance._href_definitions.get_wsd_model_attachment_href() + \
                                     "dbdrivers/" + quote(driver_file_name, safe='')
                response = requests.put(
                    content_upload_url,
                    files={'file': ('native', fdata, 'application/octet-stream', {'Expires': '0'})},
                    headers=self._client._get_headers(no_content_type=True),
                    params=self._client._params()
                )

                self._client.repository._handle_response(201, 'uploading db driver jar', response)

    def _upload_db_driver_new_api(self, path: str):
        """Upload db driver jar. Supported for IBM Cloud Pak for Data only, version 4.6.1 and above.

        :param path:  path to db driver jar
        :type path: str

        **Example**

        .. code-block:: python

            client.connections._upload_db_driver_new_api('example/path/db2jcc4.jar')

        """
        if not self._client.ICP:
            raise WMLClientError('Not supported on this environment.')

        driver_file_name = path.split('/')[-1]

        with open(path, 'rb') as fdata:
            content_upload_url = self._client.service_instance._href_definitions.get_connections_file_href(quote(driver_file_name, safe=''))
            response = requests.post(
                content_upload_url,
                data=fdata,
                headers=self._client._get_headers(content_type='application/octet-stream')
            )

            if response.status_code == 403:
                raise WMLClientError("User is missing [configure_platform] permission to upload new jar file.")

            self._client.repository._handle_response(200, 'uploading db driver jar', response, json_response=False)

    def get_db_driver_url(self, name: str): # TODO not document yet
        # """
        # Get signed db driver jar url to be used during creating of JDBC generic connection. The jar name passed as argument needs to be uploaded into system first.
        # Supported for IBM Cloud Pak for Data only, version 4.6.1 and above.
        #
        # :param name:  db driver jar name
        # :type name: str
        #
        # **Example**
        #
        # .. code-block:: python
        #
        #     client.connections.get_db_driver_url('db2jcc4.jar')
        #
        # """
        if not self._client.ICP:
            raise WMLClientError('Not supported on this environment.')

        try:
            return self.get_uploaded_db_drivers()[name]
        except WMLClientError as e:
            raise e
        except:
            raise WMLClientError(f"Driver jar with name {name} not found.")

    def sign_db_driver_url(self, jar_name: str):
        """Get signed db driver jar url to be used during creating of JDBC generic connection.
        The jar name passed as argument needs to be uploaded into system first.
        Supported for IBM Cloud Pak for Data only, version 4.0.4 and above.

        :param jar_name:  db driver jar name
        :type jar_name: str

        :return: signed db driver url
        :rtype: str

        **Example**

        .. code-block:: python

            jar_uri = client.connections.sign_db_driver_url('db2jcc4.jar')

        """
        try:
            #if self._client.ICP_46:
            #    print("Warning: This function is deprecated. Use `get_db_driver_url(name)` instead.")
            res = self.get_db_driver_url(jar_name)
            return res
        except:
            if not self._client.ICP:
                raise WMLClientError('Not supported on this environment.')

            signed_url = self._client.service_instance._href_definitions.get_wsd_model_attachment_href() + \
                         quote("dbdrivers/" + jar_name, safe='') + "/signed"
            params = self._client._params()

            params['expires_in'] = 5000

            response = requests.post(
                signed_url,
                headers=self._client._get_headers(no_content_type=True),
                params=params
            )

            self._client.repository._handle_response(201, 'signing db driver url', response, json_response=False)

            return unquote(response.headers['Location'])

    @staticmethod
    def requests_retry_session(retries=3, backoff_factor=0.3, status_forcelist=(500, 502, 503, 504, 521, 524), session=None):
        from requests.adapters import HTTPAdapter
        from requests.packages.urllib3.util.retry import Retry

        session = session or requests.Session()
        retry = Retry(
            total=retries,
            read=retries,
            connect=retries,
            backoff_factor=backoff_factor,
            status_forcelist=status_forcelist,
            raise_on_status=False
        )
        adapter = HTTPAdapter(max_retries=retry)
        session.mount('http://', adapter)
        session.mount('https://', adapter)
        return session



"""
IBM Confidential
OCO Source Materials
5737-H76, 5725-W78, 5900-A1R
(c) Copyright IBM Corp. 2020 All Rights Reserved.
The source code for this program is not published or otherwise divested of its trade secrets,
irrespective of what has been deposited with the U.S. Copyright Office.
"""
import logging

from ibmfl.exceptions import LocalTrainingException, \
    ModelUpdateException

logger = logging.getLogger(__name__)

class LocalTrainingHandler():

    def __init__(self, fl_model, data_handler, hyperparams=None, **kwargs):
        """
        Initialize LocalTrainingHandler with fl_model, data_handler

        :param fl_model: model to be trained
        :type fl_model: `model.FLModel`
        :param data_handler: data handler that will be used to obtain data
        :type data_handler: `DataHandler`
        :param hyperparams: Hyperparameters used for training.
        :type hyperparams: `dict`
        :param kwargs: Additional arguments to initialize a local training \
        handler, e.g., a crypto library object to help with encryption and \
        decryption.
        :type kwargs: `dict`
        :return None
        """
        self.fl_model = fl_model
        self.data_handler = data_handler
        self.hyperparams = hyperparams

        self.metrics_recorder = None
        self.n_completed_trains = 0
        self.n_completed_evals = 0

    def set_metrics_recorder_obj(self, metrics_recorder):
        """
        Set metrics instance variable to the input parameter. We do this because the \
        party_protocol_handler tells the local_training_handler which metrics object to use; the \
        local_training_handler can be constructed somewhere else, so we don't want to force the \
        metrics object to necessarily exist at that time.

        :param metrics_recorder: Metrics-recording object (probably empty at time this is called)
        :type metrics_recorder: `MetricsRecorder`
        """
        self.metrics_recorder = metrics_recorder

    def update_model(self, model_update):
        """
        Update local model with model updates received from FusionHandler

        :param model_update: ModelUpdate
        :type model_update: `ModelUpdate`
        """
        try:
            if model_update is not None:
                self.fl_model.update_model(model_update)
                logger.info('Local model updated.')
            else:
                logger.info('No model update was provided.')
        except Exception as ex:
            raise LocalTrainingException('No query information is provided. '+ str(ex))

    def get_train_metrics_pre(self):
        """
        Call the post-train metrics hook. This hook runs immediately before the training starts at
        each party during the routine corresponding to the TRAIN command.

        :param: None
        :return: None
        """
        if self.metrics_recorder:
            try:
                # TODO: find sound way of determining if we really want to do a pre_train (i.e.
                # "synced model") eval
                if self.metrics_recorder.compute_pre_train_eval and self.get_n_completed_trains() > 0:
                    ret = self.data_handler.get_data()
                    if ret is not None:
                        (_), test_dataset = ret
                    else:
                        test_dataset = None
                    pre_eval_results = self.fl_model.evaluate(test_dataset)
                else:
                    pre_eval_results = None
                # collect metrics specific to the model class, that the user may customize
                additional_metrics = self.fl_model.get_custom_metrics_pre()
                self.metrics_recorder.pre_train_hook(pre_eval_results, additional_metrics)
            except Exception as e:
                logger.exception(str(e))
                raise LocalTrainingException(
                    'Error occurred while running pre-train hooks')

    def get_train_metrics_post(self):
        """
        Call the post-train metrics hook. This hook runs immediately after the training finishes at
        each party during the routine corresponding to the TRAIN command.

        :param: None
        :return: None
        """
        if self.metrics_recorder:
            try:
                train_result = self.fl_model.get_train_result()
                # TODO: find sound way of determining if we really want to do a post_train (i.e.
                # "locally-trained") eval
                if self.metrics_recorder.compute_post_train_eval:
                    ret = self.data_handler.get_data()
                    if ret is not None:
                        (_), test_dataset = ret
                    else:
                        test_dataset = None
                    post_eval_results = self.fl_model.evaluate(test_dataset)
                else:
                    post_eval_results = None
                additional_metrics = self.fl_model.get_custom_metrics_post()
                self.metrics_recorder.post_train_hook(train_result, post_eval_results, additional_metrics)
            except Exception as e:
                logger.exception(str(e))
                raise LocalTrainingException(
                    'Error occurred while running post-train hooks')

    def train(self, fit_params=None):
        """
        Train locally using fl_model. At the end of training, a
        model_update with the new model information is generated and
        send through the connection.

        :param fit_params: (optional) Query instruction from aggregator
        :type fit_params: `dict`
        :return: ModelUpdate
        :rtype: `ModelUpdate`
        """
        train_data, (_) = self.data_handler.get_data()

        self.update_model(fit_params.get('model_update'))

        self.get_train_metrics_pre()

        logger.info('Local training started...')

        self.fl_model.fit_model(train_data, fit_params)

        update = self.fl_model.get_model_update()
        logger.info('Local training done, generating model update...')

        self.get_train_metrics_post()

        return update

    def save_model(self, payload=None):
        """
        Save the local model.

        :param payload: data payload received from Aggregator
        :type payload: `dict`
        :return: Status of save model request
        :rtype: `boolean`
        """
        status = False
        try:
            self.fl_model.save_model()
            status = True
        except Exception as ex:
            logger.error("Error occurred while saving local model")
            logger.exception(ex)

        return status

    def get_update_metrics_pre(self):
        """
        Call the pre-update metrics hook. This hook runs before the model update from the SYNC
        command, but after the SYNC command instruction has been received.

        :param: None
        :return: None
        """
        if self.metrics_recorder:
            try:
                self.metrics_recorder.pre_update_hook()
            except Exception as e:
                logger.exception(str(e))
                raise LocalTrainingException(
                    'Error occurred while running pre-update hooks')

    def get_update_metrics_post(self):
        """
        Call the post-update metrics hook. This hook runs after the model update from the SYNC
        command, but still during the routine corresponding to that SYNC.

        :param: None
        :return: None
        """
        if self.metrics_recorder:
            try:
                self.metrics_recorder.post_update_hook()
            except Exception as e:
                logger.exception(str(e))
                raise LocalTrainingException(
                    'Error occurred while running post-update hooks')

    def sync_model_impl(self, payload=None):
        """
        Update the local model with global ModelUpdate received
        from the Aggregator. This function is meant to be 
        overridden in base classes as opposed to sync_model, which
        contains boilerplate for exception handling and metrics.

        :param payload: data payload received from Aggregator
        :type payload: `dict`
        :return: Status of sync model request
        :rtype: `boolean`
        """
        status = False
        model_update = payload['model_update']
        status = self.fl_model.update_model(model_update)
        return status
    
    def sync_model(self, payload=None):
        """
        Update the local model with global ModelUpdate received
        from the Aggregator.

        :param payload: data payload received from Aggregator
        :type payload: `dict`
        :return: Status of sync model request
        :rtype: `boolean`
        """
        status = False
        if payload is None or 'model_update' not in payload:
            raise ModelUpdateException(
                "Invalid Model update request aggregator")

        self.get_update_metrics_pre()

        try:
            status = self.sync_model_impl(payload)
        except Exception as ex:
            logger.error("Exception occurred while sync model")
            logger.exception(ex)

        self.get_update_metrics_post()

        return status

    def eval_model(self, payload=None):
        """
        Evaluate the local model based on the local test data.

        :param payload: data payload received from Aggregator
        :type payload: `dict`
        :return: Dictionary of evaluation results
        :rtype: `dict`
        """

        (_), test_dataset = self.data_handler.get_data()
        evaluations = dict()
        try:
            evaluations = self.fl_model.evaluate(test_dataset)
            logger.info(evaluations)

        except Exception as ex:
            logger.error("Expecting the test dataset to be of type tuple. "
                         "However, test dataset is of type "
                         + str(type(test_dataset)))
            logger.exception(ex)

        return evaluations

    def get_n_completed_trains(self):
        """
        Return the number of completed executions of the TRAIN command at the party side

        :param: None
        :return: Number indicating how many TRAINs have been completed
        :rtype: `int`
        """
        return self.n_completed_trains

    def get_n_completed_evals(self):
        """
        Return the number of completed executions of the EVAL command at the party side

        :param: None
        :return: Number indicating how many EVALs have been completed
        :rtype: `int`
        """
        return self.n_completed_evals



#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2020- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

__all__ = [
    "AutoPipelinesRuns"
]


from copy import deepcopy
from typing import List, Dict, Union, Optional

from pandas import DataFrame

from ibm_watson_machine_learning.experiment.autoai.engines import WMLEngine
from ibm_watson_machine_learning.experiment.autoai.optimizers import RemoteAutoPipelines
from ibm_watson_machine_learning.preprocessing import DataJoinGraph
from ibm_watson_machine_learning.utils.autoai.utils import NextRunDetailsGenerator, get_node_and_runtime_index
from ibm_watson_machine_learning.helpers import DataConnection, S3Location, AssetLocation
from ibm_watson_machine_learning.utils.autoai.enums import ForecastingPipelineTypes
from .base_auto_pipelines_runs import BaseAutoPipelinesRuns
from ibm_watson_machine_learning.wml_client_error import ApiRequestFailure


class AutoPipelinesRuns(BaseAutoPipelinesRuns):
    """AutoPipelinesRuns class is used to work with historical Optimizer runs.

    :param engine: WMLEngine to handle WML operations
    :type engine: WMLEngine

    :param filter: filter, user can choose which runs to fetch specifying AutoPipelines name
    :type filter: str, optional
    """

    def __init__(self, engine: 'WMLEngine', filter: str = None) -> None:
        self._wml_engine: 'WMLEngine' = engine
        self.auto_pipeline_optimizer_name = filter
        self._workspace = None

    def __call__(self, *, filter: str = None) -> 'AutoPipelinesRuns':
        self.auto_pipeline_optimizer_name = filter
        return self

    def list(self) -> 'DataFrame':
        """Lists historical runs/fits with status. If user has a lot of runs stored in the WML,
        it may take long time to fetch all the information.

        :return: Pandas DataFrame with runs IDs and state
        :rtype: pandas.DataFrame
        """

        columns = ['timestamp', 'run_id', 'state', 'auto_pipeline_optimizer name']

        # note: download all runs details
        runs_details = self._wml_engine._wml_client.training.get_details(get_all=True if self.auto_pipeline_optimizer_name else False,
                                                                         limit=None if self.auto_pipeline_optimizer_name else 50,
                                                                         _internal=True)
        data = runs_details.get('resources', [])

        # note: some of the pending experiments do not have these information (checking with if statement)
        runs_pipeline_ids = [run['entity']['pipeline']['id'] for run in data if
                             run['entity'].get('pipeline', {}).get('id')]
        runs_timestamps = [run['metadata'].get('modified_at') for run in data if
                           run['entity'].get('pipeline', {}).get('id')]
        data = [run for run in data if run['entity'].get('pipeline', {}).get('id')]
        # --- end note


        def get_value(wml_pipeline_id, timestamp, run):
            try:
                pipeline_details = self._wml_engine._wml_client.pipelines.get_details(
                    pipeline_uid=wml_pipeline_id)
            except ApiRequestFailure:
                pipeline_details = {'metadata': {'name': 'Experiment data is missing...'}}

            if self.auto_pipeline_optimizer_name and pipeline_details['metadata']['name'] != self.auto_pipeline_optimizer_name:
                return None

            if not ('automl' in str(pipeline_details) or 'autoai-ts' in str(pipeline_details)) or not 'hybrid' in str(pipeline_details):
                return None

            wml_pipeline_name = pipeline_details['metadata'].get('name', 'Unknown')

            return [timestamp, run['metadata'].get('id', run['metadata'].get('guid')), run['entity']['status']['state'],
                 wml_pipeline_name]

        values = []
        for wml_pipeline_id, timestamp, run in zip(runs_pipeline_ids, runs_timestamps, data):
            v = get_value(wml_pipeline_id, timestamp, run)
            if v:
                values.append(v)
            if len(values) >= 50:
                break

        self.auto_pipeline_optimizer_name = None
        runs = DataFrame(data=values, columns=columns)
        return runs.sort_values(by=["timestamp"], ascending=False)[:50]

    def get_params(self, run_id: str = None) -> dict:
        """Get executed optimizers configs parameters based on the run_id.

        :param run_id: ID of the fit/run, if not specified, latest is taken
        :type run_id: str, optional

        :return: optimizer configuration parameters
        :rtype: dict

        **Example**

        .. code-block:: python

            from ibm_watson_machine_learning.experiment import AutoAI
            experiment = AutoAI(credentials, ...)

            experiment.runs.get_params(run_id='02bab973-ae83-4283-9d73-87b9fd462d35')
            experiment.runs.get_params()

            # Result:
            # {
            #     'name': 'test name',
            #     'desc': 'test description',
            #     'prediction_type': 'classification',
            #     'prediction_column': 'y',
            #     'scoring': 'roc_auc',
            #     'holdout_size': 0.1,
            #     'max_num_daub_ensembles': 1
            # }
        """

        if run_id is None:
            optimizer_id = self._wml_engine._wml_client.training.get_details(
                limit=1, _internal=True
            ).get('resources')[0]['entity']['pipeline']['id']

        else:
            optimizer_id = self._wml_engine._wml_client.training.get_details(
                training_uid=run_id, _internal=True
            ).get('entity')['pipeline']['id']

        optimizer_config = self._wml_engine._wml_client.pipelines.get_details(pipeline_uid=optimizer_id)

        # note: if experiment has more than 1 node (e.g. OBM + KB), we need to find which one is KB
        kb_node_number, kb_runtime_number = get_node_and_runtime_index(node_name='kb',
                                                                       optimizer_config=optimizer_config)
        # --- end note

        # note: try to find obm node
        obm_node_number, obm_runtime_number = get_node_and_runtime_index(node_name='obm',
                                                                         optimizer_config=optimizer_config)
        # --- end note

        # note: try to find ts node
        ts_node_number, ts_runtime_number = get_node_and_runtime_index(node_name='ts',
                                                                       optimizer_config=optimizer_config)
        # --- end note

        # note: try to find tsad node
        tsad_node_number, tsad_runtime_number = get_node_and_runtime_index(node_name='tsad',
                                                                           optimizer_config=optimizer_config)
        # --- end note

        try:
            name = optimizer_config['entity']['name']
            description = optimizer_config['entity'].get('description', '')

        except KeyError:
            name = optimizer_config['metadata']['name']
            description = optimizer_config['metadata'].get('description', '')

        # note: check if not only data preprocessing experiment
        if kb_node_number is not None:
            kb_parameters = optimizer_config['entity']['document']['pipelines'][0]['nodes'][kb_node_number][
                'parameters']['optimization']
            kb_wml_data = optimizer_config['entity']['document']['runtimes'][kb_runtime_number]['app_data']['wml_data']

            csv_separator = optimizer_config['entity']['document']['pipelines'][0]['nodes'][kb_node_number][
                'parameters'].get('input_file_separator', ',')
            excel_sheet = optimizer_config['entity']['document']['pipelines'][0]['nodes'][kb_node_number][
                'parameters'].get('excel_sheet', None)
            encoding = optimizer_config['entity']['document']['pipelines'][0]['nodes'][kb_node_number][
                'parameters'].get('encoding', 'utf-8')
            drop_duplicates = optimizer_config['entity']['document']['pipelines'][0]['nodes'][kb_node_number][
                'parameters'].get('drop_duplicates', True)

            params = {
                'name': name,
                'desc': description,
                'prediction_type': kb_parameters['learning_type'],
                'prediction_column': kb_parameters.get('label'),
                'prediction_columns': kb_parameters.get('target_columns'),
                'timestamp_column_name': kb_parameters.get('timestamp_column'),
                'holdout_size': kb_parameters.get('holdout_param'),
                'max_num_daub_ensembles': kb_parameters.get('max_num_daub_ensembles'),
                't_shirt_size': kb_wml_data['hardware_spec']['id'],
                'include_only_estimators': kb_parameters.get('daub_include_only_estimators'),
                'cognito_transform_names': kb_parameters.get('cognito_transform_names'),
                'train_sample_rows_test_size': kb_parameters.get('train_sample_rows_test_size'),
                'text_processing': kb_parameters.get('text_processing_flag'),
                'train_sample_columns_index_list': kb_parameters.get('train_sample_columns_index_list'),
                'daub_give_priority_to_runtime': kb_parameters.get('daub_runtime_ranking_power'),
                'positive label': kb_parameters.get('daub_runtime_ranking_power'),
                'drop_duplicates': drop_duplicates,
                'csv_separator': csv_separator,
                'excel_sheet': excel_sheet,
                'encoding': encoding,
                'retrain_on_holdout': kb_parameters.get('retrain_on_holdout')
            }

            if kb_parameters.get('train_sample_rows_test_size'):
                params['train_sample_rows_test_size'] = kb_parameters['train_sample_rows_test_size']
            if kb_parameters.get('scorer_for_ranking'):
                params['scoring'] = kb_parameters['scorer_for_ranking']
            if kb_parameters.get('text_processing_options') and kb_parameters['text_processing_options'].get(
                    'word2vec'):
                params['word2vec_feature_number'] = kb_parameters['text_processing_options']['word2vec'].get(
                    'output_dim')
            if kb_parameters.get('fairness_info'):
                params['fairness_info'] = kb_parameters.get('fairness_info')
        elif ts_node_number is not None:
            ts_parameters = optimizer_config['entity']['document']['pipelines'][0]['nodes'][ts_node_number][
                'parameters']['optimization']
            ts_wml_data = optimizer_config['entity']['document']['runtimes'][ts_node_number]['app_data']['wml_data']

            csv_separator = optimizer_config['entity']['document']['pipelines'][0]['nodes'][ts_node_number][
                'parameters'].get('input_file_separator', ',')
            encoding = optimizer_config['entity']['document']['pipelines'][0]['nodes'][ts_node_number][
                'parameters'].get('encoding', 'utf-8')

            if ts_parameters.get('pipeline_type') == 'customized':
                pipeline_types = ts_parameters.get('customized_pipelines')
                pipeline_types = [ForecastingPipelineTypes(p) for p in pipeline_types]
            elif ts_parameters.get('pipeline_type') == 'exogenous':
                pipeline_types = ForecastingPipelineTypes.get_exogenous()
            elif ts_parameters.get('pipeline_type') == 'non_exogenous':
                pipeline_types = ForecastingPipelineTypes.get_non_exogenous()
            elif ts_parameters.get('pipeline_type') == 'all':
                pipeline_types = [l for l in ForecastingPipelineTypes]
            else:
                pipeline_types = None

            params = {
                'name': name,
                'desc': description,
                'prediction_type': ts_parameters['learning_type'],
                'prediction_column': ts_parameters.get('label'),
                'prediction_columns': ts_parameters.get('target_columns'),
                'timestamp_column_name': ts_parameters.get('timestamp_column'),
                'holdout_size': ts_parameters.get('holdout_param', 20),
                'max_num_daub_ensembles': ts_parameters.get('max_num_daub_ensembles', 3),
                'backtest_gap_length': ts_parameters.get('gap_len'),
                'backtest_num': ts_parameters.get('num_backtest'),
                'forecast_window': ts_parameters.get('prediction_horizon'),
                'include_only_estimators': ts_parameters.get('include_only_estimators'),
                'lookback_window': ts_parameters.get('lookback_window'),
                't_shirt_size': ts_wml_data['hardware_spec']['id'],
                'csv_separator': csv_separator,
                'encoding': encoding,
                'feature_columns': ts_parameters.get('feature_columns'),
                'supporting_features_at_forecast': ts_parameters.get('future_exogenous_available'),
                'pipeline_types': pipeline_types,
                'retrain_on_holdout': ts_parameters.get('retrain_on_holdout', True),
            }
        elif tsad_node_number is not None:
            tsad_parameters = optimizer_config['entity']['document']['pipelines'][0]['nodes'][tsad_node_number][
                'parameters']['optimization']
            tsad_wml_data = optimizer_config['entity']['document']['runtimes'][tsad_node_number]['app_data']['wml_data']

            csv_separator = optimizer_config['entity']['document']['pipelines'][0]['nodes'][tsad_node_number][
                'parameters'].get('input_file_separator', ',')
            encoding = optimizer_config['entity']['document']['pipelines'][0]['nodes'][tsad_node_number][
                'parameters'].get('encoding', 'utf-8')

            params = {
                'name': name,
                'desc': description,
                'prediction_type': tsad_parameters.get('learning_type'),
                'feature_columns': tsad_parameters.get('feature_columns'),
                'prediction_column': tsad_parameters.get('label'),
                'prediction_columns': tsad_parameters.get('target_columns'),
                'timestamp_column_name': tsad_parameters.get('timestamp_column'),
                'holdout_size': tsad_parameters.get('holdout_param', 0.2),
                'max_num_daub_ensembles': tsad_parameters.get('max_num_pipelines', 3),
                'include_only_estimators': tsad_parameters.get('daub_include_only_estimators'),
                't_shirt_size': tsad_wml_data['hardware_spec']['id'],
                'csv_separator': csv_separator,
                'encoding': encoding,
                'pipeline_types': tsad_parameters.get('pipelines'),
                'retrain_on_holdout': tsad_parameters.get('retrain_on_holdout', True),
                'scoring': tsad_parameters.get('evaluation_metric', 'average_precision'),
                # 'confidence_level': tsad_parameters.get('confidence_level', 0.95),
            }
        else:
            params = {
                'name': name,
                'desc': description,
                'prediction_type': None,
                'prediction_column': None,
                'prediction_columns': None,
                'timestamp_column_name': None,
                'scoring': None,
                'data_join_only': True
            }
        # --- end note

        # note: rebuild DataJoinGraph object from historical run
        if obm_node_number is not None:
            obm_parameters = optimizer_config['entity']['document']['pipelines'][0]['nodes'][obm_node_number]
            data_join_graph = DataJoinGraph._from_dict(_dict=obm_parameters)

            obm_wml_data = optimizer_config['entity']['document']['runtimes'][obm_runtime_number]['app_data'][
                'wml_data']
            # need to trim X-spark t-shirt name only to first letter
            t_name = obm_wml_data['hardware_spec']['name']
            data_join_graph.t_shirt_size = t_name if len(t_name) == 1 else t_name[0]

            params['data_join_graph'] = data_join_graph
        # --- end note

        return params

    def get_run_details(self, run_id: str = None) -> dict:
        """Get run details. If run_id is not supplied, last run will be taken.

        :param run_id: ID of the fit/run
        :type run_id: str, optional

        :return: run configuration parameters
        :rtype: dict

        **Example**

        .. code-block:: python

            from ibm_watson_machine_learning.experiment import AutoAI
            experiment = AutoAI(credentials, ...)

            experiment.runs.get_run_details(run_id='02bab973-ae83-4283-9d73-87b9fd462d35')
            experiment.runs.get_run_details()
        """
        if run_id is None:
            details = self._wml_engine._wml_client.training.get_details(limit=1, _internal=True).get('resources')[0]

        else:
            details = self._wml_engine._wml_client.training.get_details(training_uid=run_id, _internal=True)

        if details['entity']['status'].get('metrics', False):
            del details['entity']['status']['metrics']
            return details
        else:
            return details

    def get_optimizer(self,
                      run_id: Optional[str] = None,
                      metadata: Dict[str, Union[List['DataConnection'], 'DataConnection', str, int]] = None
                      ) -> 'RemoteAutoPipelines':
        """Create instance of AutoPipelinesRuns with all computed pipelines computed by AutoAi on WML.

        :param run_id: ID of the fit/run
        :type run_id: str, optional

        :param metadata: option to pass information about COS data reference or WSD (auto-gen notebook)
        :type metadata: dict, optional

        :return: optimizer object
        :rtype: AutoPipelinesRuns class instance

        **Example**

        .. code-block:: python

            from ibm_watson_machine_learning.experiment import AutoAI
            experiment = AutoAI(credentials, ...)

            historical_optimizer = experiment.runs.get_optimizer(run_id='02bab973-ae83-4283-9d73-87b9fd462d35')
        """
        # note: normal scenario
        if metadata is None:
            optimizer_parameters = self.get_params(run_id=run_id)

            remote_pipeline_optimizer = RemoteAutoPipelines(**optimizer_parameters, engine=self._wml_engine)

            remote_pipeline_optimizer._engine._current_run_id = run_id
            remote_pipeline_optimizer._workspace = self._workspace

            return remote_pipeline_optimizer
        # --- end note

        # note: WSD / Cloud auto-gen notebook scenario (when user provides his WMLS credentials)
        else:
            from ibm_watson_machine_learning.experiment import AutoAI
            training_result_reference = metadata.get('training_result_reference')

            # note: check for cloud
            if isinstance(training_result_reference.location, (S3Location, AssetLocation)):
                run_id = training_result_reference.location._training_status.split('/')[-2]

            # WMLS
            else:
                run_id = training_result_reference.location.path.split('/')[-3]

            # note: CP4D notebook scenario
            if self._wml_engine is not None:
                return AutoAI(self._workspace).runs.get_optimizer(run_id)

            # note WSD
            else:
                return AutoAI(self._workspace.wml_credentials,
                              space_id=self._workspace.space_id
                              ).runs.get_optimizer(run_id)
        # --- end note

    def get_data_connections(self, run_id: str) -> List['DataConnection']:
        """Create DataConnection objects for further user usage
            (eg. to handle data storage connection or to recreate autoai holdout split).

        :param run_id: ID of the historical fit/run
        :type run_id: str

        :return: list of DataConnections with populated optimizer parameters
        :rtype: list['DataConnection']

        **Example**

        .. code-block:: python

            from ibm_watson_machine_learning.experiment import AutoAI
            experiment = AutoAI(credentials, ...)

            data_connections = experiment.runs.get_data_connections(run_id='02bab973-ae83-4283-9d73-87b9fd462d35')
        """
        optimizer_parameters = self.get_params(run_id=run_id)
        training_data_references = self.get_run_details(run_id=run_id)['entity']['training_data_references']

        data_connections = [
            DataConnection._from_dict(_dict=data_connection) for data_connection in training_data_references]

        for data_connection in data_connections:  # note: populate DataConnections with optimizer params
            data_connection.auto_pipeline_params = deepcopy(optimizer_parameters)
            data_connection.set_client(self._wml_engine._wml_client)
            data_connection._run_id = run_id

        return data_connections



"""
IBM Confidential
OCO Source Materials
5737-H76, 5725-W78, 5900-A1R
(c) Copyright IBM Corp. 2020 All Rights Reserved.
The source code for this program is not published or otherwise divested of its trade secrets,
irrespective of what has been deposited with the U.S. Copyright Office.
"""
import os
import pathlib
from environs import Env

env = Env()
env.read_env()

# export FL_WORKING_DIR=/home/username/
# export FL_MODEL_DIR=/home/username/model


def create_dir(path):
    pathlib.Path(path).mkdir(parents=True, exist_ok=True)


with env.prefixed("FL_"):
    working_directory = env("WORKING_DIR", os.getcwd())
    create_dir(working_directory)

    model_directory = env("MODEL_DIR", working_directory)
    create_dir(model_directory)



from __future__ import print_function
import numpy as np
from scipy.optimize import linear_sum_assignment


def build_init(hungarian_weights, assignments, j):
    """
    Create local network weights from the matched global network weights

    :param hungarian_weights: global network weights
    :type hungarian_weights: `numpy.ndarray`
    :param assignments: assignment matrix mapping local to global neurons
    :type  assignments: `numpy.ndarray`
    :param j: network index for which updated local weights are required
    :type j: `int`
    :return: local network weights.
    :rtype: `list` of `list`
    """
    batch_init = []
    num_assignments = len(assignments)

    if len(hungarian_weights) == 4:
        batch_init.append(hungarian_weights[0][:, assignments[0][j]])
        batch_init.append(hungarian_weights[1][assignments[0][j]])
        batch_init.append(hungarian_weights[2][assignments[0][j]])
        batch_init.append(hungarian_weights[3])
        return batch_init

    for c in range(num_assignments):
        if c == 0:
            batch_init.append(hungarian_weights[c][:, assignments[c][j]])
            batch_init.append(hungarian_weights[c + 1][assignments[c][j]])
        else:
            batch_init.append(hungarian_weights[2 * c]
                              [assignments[c - 1][j]]
                              [:, assignments[c][j]])
            batch_init.append(hungarian_weights[2 * c + 1]
                              [assignments[c][j]])
            if c == num_assignments - 1:
                batch_init.append(hungarian_weights[2 * c + 2]
                                  [assignments[c][j]])
                batch_init.append(hungarian_weights[-1])
                return batch_init


def row_param_cost(global_weights, weights_j_l, global_sigmas, sigma_inv_j):
    """
    Computes the cost defined in equation 6 of PFNM paper
    :param global_weights:
    :type global_weights: `numpy.ndarray`
    :param weights_j_l:
    :type weights_j_l: `numpy.ndarray`
    :param global_sigmas:
    :type global_sigmas: `numpy.ndarray`
    :param sigma_inv_j:
    :type sigma_inv_j: `numpy.ndarray`
    :return Cost value
    :rtype: `numpy.ndarray`
    """

    match_norms = ((weights_j_l + global_weights) ** 2 /
                   (sigma_inv_j + global_sigmas)).sum(axis=1) - \
                  (global_weights ** 2 / global_sigmas).sum(axis=1)
    return match_norms


def compute_cost(global_weights, weights_j, global_sigmas, sigma_inv_j,
                 prior_mean_norm, prior_inv_sigma, popularity_counts, gamma, J):
    """
    Computes full cost to be used by Hungarian algorithm. Refer Equation (8) in the PFNM paper
    :param global_weights: weight matrix
    :type global_weights: `numpy.ndarray`
    :param weights_j: Jth network weights
    :type weights_j: `numpy.ndarray`
    :param global_sigmas: global sigma values
    :type global_sigmas: `numpy.ndarray`
    :param sigma_inv_j: Inverse sigma values for jth network
    :type sigma_inv_j: `numpy.ndarray`
    :param prior_mean_norm: Prior mean values
    :type prior_mean_norm: `numpy.ndarray`
    :param prior_inv_sigma: Prior inverse sigma  values
    :type prior_inv_sigma: `numpy.ndarray`
    :param popularity_counts: popularity count values
    :type popularity_counts: `list`
    :param gamma: Gamma value
    :type gamma: `float`
    :param J: number of clients
    :type J: `int`
    :return: Cost value
    :rtype: `numpy.ndarray`
    """

    Lj = weights_j.shape[0]
    counts = np.minimum(np.array(popularity_counts), 10)
    param_cost = np.array(
        [
            row_param_cost(global_weights, weights_j[l], global_sigmas, sigma_inv_j)
            for l in range(Lj)
        ]
    )
    param_cost += np.log(counts / (J - counts))

    # Nonparametric cost
    L = global_weights.shape[0]
    max_added = min(Lj, max(700 - L, 1))

    nonparam_cost = np.outer((((weights_j + prior_mean_norm) ** 2 /
                               (prior_inv_sigma + sigma_inv_j)).sum(axis=1)
                              - (prior_mean_norm ** 2 /
                                 prior_inv_sigma).sum()),
                             np.ones(max_added))
    cost_pois = 2 * np.log(np.arange(1, max_added + 1))
    nonparam_cost -= cost_pois
    nonparam_cost += 2 * np.log(gamma / J)

    full_cost = np.hstack((param_cost, nonparam_cost))

    return full_cost


def matching_upd_j(weights_j, global_weights, sigma_inv_j, global_sigmas,
                   prior_mean_norm, prior_inv_sigma, popularity_counts, gamma, J):
    """
    Computes cost [Equation 8] and solves the linear assignment problem
    using hungarian algo
    :param weights_j: Jth network weights
    :type weights_j: `numpy.ndarray`
    :param global_weights: global network weights
    :type global_weights: `numpy.ndarray`
    :param sigma_inv_j: Inverse sigma values for jth network
    :type sigma_inv_j: `numpy.ndarray`
    :param global_sigmas: Global sigma values
    :type global_sigmas: `numpy.ndarray`
    :param prior_mean_norm: Prior mean values
    :type prior_mean_norm: `numpy.ndarray`
    :param prior_inv_sigma: Prior inverse sigma values
    :type prior_inv_sigma: `numpy.ndarray`
    :param popularity_counts: Popularity counts
    :type popularity_counts: `list`
    :param gamma: Gamma values
    :type gamma: `float`
    :param J: number of clients
    :type J: `int`
    :return a tuple of global weights, sigma values, popularity counts, and assignment values
    :rtype: `tuple`
    """

    L = global_weights.shape[0]

    full_cost = compute_cost(global_weights, weights_j, global_sigmas, sigma_inv_j,
                             prior_mean_norm, prior_inv_sigma, popularity_counts, gamma, J)

    row_ind, col_ind = linear_sum_assignment(-full_cost)

    assignment_j = []
    new_L = L

    for l, i in zip(row_ind, col_ind):
        if i < L:
            popularity_counts[i] += 1
            assignment_j.append(i)
            global_weights[i] += weights_j[l]
            global_sigmas[i] += sigma_inv_j
        else:  # new neuron
            popularity_counts += [1]
            assignment_j.append(new_L)
            new_L += 1
            global_weights = np.vstack((global_weights,
                                        prior_mean_norm + weights_j[l]))
            global_sigmas = np.vstack((global_sigmas,
                                       prior_inv_sigma + sigma_inv_j))

    return global_weights, global_sigmas, popularity_counts, assignment_j


def patch_weights(w_j, L_next, assignment_j_c):
    """
    Patch weights for different layers of the network together into a
    single array of weights
    :param w_j: Jth network weights
    :type w_j: `numpy.ndarray`
    :param L_next: next layer dimensions
    :type L_next: `int`
    :param assignment_j_c: assignment matrix
    :type assignment_j_c: `numpy.ndarray`
    :return: patched weight matrix
    :rtype: `numpy.ndarray`
    """
    if assignment_j_c is None:
        return w_j
    new_w_j = np.zeros((w_j.shape[0], L_next))
    new_w_j[:, assignment_j_c] = w_j
    return new_w_j


def process_softmax_bias(batch_weights, last_layer_const, sigma, sigma0):
    """
    Process the weights of the final classification layer. Since the final
    layer weights are directly associated with classes, these can be
    merged directly without finding an assignment matrix that's required
    for hidden layers
    :param batch_weights: weight matrix
    :type batch_weights: `numpy.ndarray`
    :param last_layer_const: last layer scaling factor
    :type last_layer_const: `list`
    :param sigma: sigma value
    :type sigma: `float`
    :param sigma0: sigma0 value
    :type sigma0: `float`
    :return: a tuple of new softmax bias values and sigma inverse values
    :rtype: `tuple`
    """

    J = len(batch_weights)
    sigma_bias = sigma
    sigma0_bias = sigma0
    mu0_bias = 0.1
    softmax_bias = [batch_weights[j][-1] for j in range(J)]
    softmax_inv_sigma = [s / sigma_bias for s in last_layer_const]
    softmax_bias = sum(
        [b * s for b, s in zip(softmax_bias, softmax_inv_sigma)]
    ) + mu0_bias / sigma0_bias
    softmax_inv_sigma = 1 / sigma0_bias + sum(softmax_inv_sigma)
    return softmax_bias, softmax_inv_sigma


def init_from_assignments(batch_weights_norm, sigma_inv_layer, prior_mean_norm,
                          sigma_inv_prior, assignment):
    """
    Initialize a global network using the assignment matrix of
    local networks found by hungarian algorithm
    :param batch_weights_norm: network weights
    :type batch_weights_norm: `numpy.ndarray`
    :param sigma_inv_layer: inverse sigma values for layer
    :type sigma_inv_layer: `list`
    :param prior_mean_norm: prior mean values
    :type prior_mean_norm: `numpy.ndarray`
    :param sigma_inv_prior: prior sigma inverse values
    :type sigma_inv_prior: `numpy.ndarray`
    :param assignment: assignment matrix
    :type assignment: `numpy.ndarray`
    :return: a tuple of popularity counts, initialized weights, and global sigma
    :rtype: `tuple`
    """

    L = int(max([max(a_j) for a_j in assignment])) + 1
    popularity_counts = [0] * L

    global_weights = np.outer(np.ones(L), prior_mean_norm)
    global_sigmas = np.outer(np.ones(L), sigma_inv_prior)

    for j, a_j in enumerate(assignment):
        for l, i in enumerate(a_j):
            popularity_counts[i] += 1
            global_weights[i] += batch_weights_norm[j][l]
            global_sigmas[i] += sigma_inv_layer[j]

    return popularity_counts, global_weights, global_sigmas





#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2017- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

from __future__ import print_function
import ibm_watson_machine_learning._wrappers.requests as requests
import json
import re
import time
import copy
from ibm_watson_machine_learning.wml_client_error import MissingValue, WMLClientError, MissingMetaProp
from ibm_watson_machine_learning.href_definitions import is_uid
from ibm_watson_machine_learning.wml_resource import WMLResource
from multiprocessing import Pool
from ibm_watson_machine_learning.utils import EXPERIMENT_DETAILS_TYPE
from ibm_watson_machine_learning.hpo import HPOParameter, HPOMethodParam
from ibm_watson_machine_learning.metanames import ExperimentMetaNames
from ibm_watson_machine_learning.href_definitions import API_VERSION, SPACES, PIPELINES, LIBRARIES, EXPERIMENTS, RUNTIMES, DEPLOYMENTS

_DEFAULT_LIST_LENGTH = 50

class Experiments(WMLResource):
    """Run new experiment."""

    ConfigurationMetaNames = ExperimentMetaNames()
    """MetaNames for experiments creation."""

    @staticmethod
    def _HPOParameter(name, values=None, max=None, min=None, step=None):
        return HPOParameter(name, values, max, min, step)

    @staticmethod
    def _HPOMethodParam(name=None, value=None):
        return HPOMethodParam(name, value)

    def __init__(self, client):
        WMLResource.__init__(self, __name__, client)
        self._experiments_uids_cache = {}

    def store(self, meta_props):
        """Create an experiment.

        :param meta_props: meta data of the experiment configuration. To see available meta names use:

            .. code-block:: python

                client.experiments.ConfigurationMetaNames.get()

        :type meta_props: dict

        :return: stored experiment metadata
        :rtype: dict

        **Example**

        .. code-block:: python

            metadata = {
                client.experiments.ConfigurationMetaNames.NAME: 'my_experiment',
                client.experiments.ConfigurationMetaNames.EVALUATION_METRICS: ['accuracy'],
                client.experiments.ConfigurationMetaNames.TRAINING_REFERENCES: [
                    {'pipeline': {'href': pipeline_href_1}},
                    {'pipeline': {'href':pipeline_href_2}}
                ]
            }
            experiment_details = client.experiments.store(meta_props=metadata)
            experiment_href = client.experiments.get_href(experiment_details)

        """
        WMLResource._chk_and_block_create_update_for_python36(self)
        ##For CP4D, check if either spce or project ID is set
        self._client._check_if_either_is_set()

        metaProps = self.ConfigurationMetaNames._generate_resource_metadata(meta_props)
        ##Check if default space is set
        if self._client.CAMS and not self._client.ICP_PLATFORM_SPACES:
            if self._client.default_space_id is not None:
                metaProps['space'] = {'href': "/v4/spaces/"+self._client.default_space_id}
            elif self._client.default_project_id is not None:
                metaProps['project'] = {'href': "/v2/projects/"+self._client.default_project_id}
            else:
                raise WMLClientError("It is mandatory to set the space/project id. Use client.set.default_space(<SPACE_UID>)/client.set.default_project(<PROJECT_UID>) to proceed.")

        if self._client.CLOUD_PLATFORM_SPACES or self._client.ICP_PLATFORM_SPACES:
            if self._client.default_space_id is not None:
                metaProps['space_id'] = self._client.default_space_id
            elif self._client.default_project_id is not None:
                metaProps['project_id'] = self._client.default_project_id
            else:
                raise WMLClientError(
                    "It is mandatory to set the space/project id. Use client.set.default_space(<SPACE_UID>)/client.set.default_project(<PROJECT_UID>) to proceed.")
            self._validate_meta_prop(meta_props, self.ConfigurationMetaNames.NAME, str, True)
        else:
            self.ConfigurationMetaNames._validate(meta_props)

        if self._client.CLOUD_PLATFORM_SPACES or self._client.ICP_PLATFORM_SPACES:
            response_experiment_post = requests.post(
                self._client.service_instance._href_definitions.get_experiments_href(),
                params=self._client._params(skip_for_create=True),
                json=metaProps,
                headers=self._client._get_headers()
            )
        else:
            response_experiment_post = requests.post(
                self._client.service_instance._href_definitions.get_experiments_href(),
                json=metaProps,
                headers=self._client._get_headers()
            )

        return self._handle_response(201, u'saving experiment', response_experiment_post)

    def update(self, experiment_uid, changes):
        """Updates existing experiment metadata.

        :param experiment_uid: UID of experiment which definition should be updated
        :type experiment_uid: str
        :param changes: elements which should be changed, where keys are ConfigurationMetaNames
        :type changes: dict

        :return: metadata of updated experiment
        :rtype: dict

        **Example**

        .. code-block:: python

            metadata = {
                client.experiments.ConfigurationMetaNames.NAME: "updated_exp"
            }
            exp_details = client.experiments.update(experiment_uid, changes=metadata)

        """
        WMLResource._chk_and_block_create_update_for_python36(self)
        ##For CP4D, check if either spce or project ID is set
        self._client._check_if_either_is_set()

        self._validate_type(experiment_uid, u'experiment_uid', str, True)
        self._validate_type(changes, u'changes', dict, True)

        details = self._client.repository.get_details(experiment_uid)

        patch_payload = self.ConfigurationMetaNames._generate_patch_payload(details['entity'], changes, with_validation=True)

        url = self._client.service_instance._href_definitions.get_experiment_href(experiment_uid)
        response = requests.patch(url, json=patch_payload, params = self._client._params(),headers=self._client._get_headers())
        updated_details = self._handle_response(200, u'experiment patch', response)

        return updated_details

    def get_details(self, experiment_uid=None, limit=None, asynchronous=False, get_all=False):
        """Get metadata of experiment(s). If no experiment UID is specified all experiments metadata is returned.

        :param experiment_uid:  UID of experiment
        :type experiment_uid: str, optional
        :param limit: limit number of fetched records
        :type limit: int, optional
        :param asynchronous: if `True`, it will work as a generator
        :type asynchronous: bool, optional
        :param get_all: if `True`, it will get all entries in 'limited' chunks
        :type get_all: bool, optional

        :return: experiment(s) metadata
        :rtype: dict (if UID is not None) or {"resources": [dict]} (if UID is None)

        **Example**

        .. code-block:: python

            experiment_details = client.experiments.get_details(experiment_uid)
            experiment_details = client.experiments.get_details()
            experiment_details = client.experiments.get_details(limit=100)
            experiment_details = client.experiments.get_details(limit=100, get_all=True)
            experiment_details = []
            for entry in client.experiments.get_details(limit=100, asynchronous=True, get_all=True):
                experiment_details.extend(entry)

         """
        ##For CP4D, check if either spce or project ID is set
        self._client._check_if_either_is_set()
        url = self._client.service_instance._href_definitions.get_experiments_href()

        if experiment_uid is None:
            return self._get_artifact_details(url, experiment_uid, limit, 'experiment',
                                              _async=asynchronous, _all=get_all)

        else:
            return self._get_artifact_details(url, experiment_uid, limit, 'experiment')

    @staticmethod
    def get_uid(experiment_details):
        """Get Unique Id of stored experiment.

        :param experiment_details: metadata of the stored experiment
        :type experiment_details: dict

        :return: Unique Id of stored experiment
        :rtype: str

        **Example**

        .. code-block:: python

            experiment_details = client.experiments.get_details(experiment_uid)
            experiment_uid = client.experiments.get_uid(experiment_details)

        """
        Experiments._validate_type(experiment_details, u'experiment_details', object, True)
        if 'id' not in experiment_details[u'metadata']:
            Experiments._validate_type_of_details(experiment_details, EXPERIMENT_DETAILS_TYPE)

        return WMLResource._get_required_element_from_dict(experiment_details, u'experiment_details',
                                                           [u'metadata', u'id'])

    @staticmethod
    def get_id(experiment_details):
        """Get Unique Id of stored experiment.

        :param experiment_details: metadata of the stored experiment
        :type experiment_details: dict

        :return: Unique Id of stored experiment
        :rtype: str

        **Example**

        .. code-block:: python

            experiment_details = client.experiments.get_details(experiment_id)
            experiment_uid = client.experiments.get_id(experiment_details)

        """
        Experiments._validate_type(experiment_details, u'experiment_details', object, True)
        if 'id' not in experiment_details[u'metadata']:
            Experiments._validate_type_of_details(experiment_details, EXPERIMENT_DETAILS_TYPE)

        return WMLResource._get_required_element_from_dict(experiment_details, u'experiment_details',
                                                           [u'metadata', u'id'])

    @staticmethod
    def get_href(experiment_details):
        """Get href of stored experiment.

        :param experiment_details: metadata of the stored experiment
        :type experiment_details: dict

        :return: href of stored experiment
        :rtype: str

        **Example**

        .. code-block:: python

            experiment_details = client.experiments.get_details(experiment_uid)
            experiment_href = client.experiments.get_href(experiment_details)

        """
        Experiments._validate_type(experiment_details, u'experiment_details', object, True)
        if 'href' in experiment_details['metadata']:
            Experiments._validate_type_of_details(experiment_details, EXPERIMENT_DETAILS_TYPE)

            return WMLResource._get_required_element_from_dict(experiment_details, u'experiment_details',
                                                           [u'metadata', u'href'])
        else:
            experiment_id = WMLResource._get_required_element_from_dict(experiment_details, u'experiment_details',
                                                                     [u'metadata', u'id'])
            return "/ml/v4/experiments/" + experiment_id

    def delete(self, experiment_uid):
        """Delete a stored experiment.

        :param experiment_uid: Unique Id of the stored experiment
        :type experiment_uid: str

        :return: status ("SUCCESS" or "FAILED")
        :rtype: str

        **Example**

        .. code-block:: python

            client.experiments.delete(experiment_uid)

        """
        ##For CP4D, check if either spce or project ID is set
        self._client._check_if_either_is_set()
        Experiments._validate_type(experiment_uid, u'experiment_uid', str, True)

        url = self._client.service_instance._href_definitions.get_experiment_href(experiment_uid)
        response = requests.delete(url, params = self._client._params(),headers=self._client._get_headers())

        return self._handle_response(204, u'experiment deletion', response, False)

    def list(self, limit=None):
        """Print stored experiments in a table format.
        If limit is set to None there will be only first 50 records shown.

        :param limit: limit number of fetched records
        :type limit: int, optional

        **Example**

        .. code-block:: python

            client.experiments.list()

        """
        ##For CP4D, check if either spce or project ID is set
        self._client._check_if_either_is_set()
        experiment_resources = self.get_details(limit=limit)[u'resources']
        header_list = [u'GUID', u'NAME', u'CREATED']
        if self._client.CLOUD_PLATFORM_SPACES or self._client.ICP_PLATFORM_SPACES:
            experiment_values = [(m[u'metadata'][u'id'], m[u'metadata'][u'name'], m[u'metadata'][u'created_at']) for m in
                                 experiment_resources]
            header_list = [u'ID', u'NAME', u'CREATED']
        else:
            experiment_values = [(m[u'metadata'][u'id'], m[u'entity'][u'name'], m[u'metadata'][u'created_at']) for m in
                                 experiment_resources]

        self._list(experiment_values, header_list, limit, _DEFAULT_LIST_LENGTH)

    def create_revision(self, experiment_id):
        """Create a new experiment revision.

        :param experiment_id: Unique Id of the stored experiment
        :type experiment_id: str

        :return: stored experiment new revision details
        :rtype: dict

        **Example**

        .. code-block:: python

            experiment_revision_artifact = client.experiments.create_revision(experiment_id)

        """
        WMLResource._chk_and_block_create_update_for_python36(self)

        ##For CP4D, check if either spce or project ID is set
        self._client._check_if_either_is_set()
        Experiments._validate_type(experiment_id, u'experiment_id', str, True)

        if not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES:
            raise WMLClientError(
                u'Revision support is not there in this WML server. It is supported only from 3.1.0 onwards.')
        else:
            url =self._client.service_instance._href_definitions.get_experiments_href()
            return self._create_revision_artifact(url, experiment_id, 'experiments')

    def get_revision_details(self, experiment_uid, rev_uid):
        """Get metadata of stored experiments revisions.

        :param experiment_uid: stored experiment UID
        :type experiment_uid: str

        :param rev_uid: rev_id number of experiment
        :type rev_uid: int

        :return: stored experiment revision metadata
        :rtype: dict

        Example:

        .. code-block:: python

            experiment_details = client.experiments.get_revision_details(experiment_uid, rev_id)

         """
        self._client._check_if_either_is_set()
        Experiments._validate_type(experiment_uid, u'experiment_uid', str, True)
        Experiments._validate_type(rev_uid, u'rev_uid', int, True)
        if not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES:
            raise WMLClientError(
                'Not supported. Revisions APIs are supported only for IBM Cloud Pak® for Data 3.0 and above.')
        else:
            url = self._client.service_instance._href_definitions.get_experiment_href(experiment_uid)
            return self._get_with_or_without_limit(url, limit=None, op_name="experiments",
                                                   summary=None, pre_defined=None, revision=rev_uid)

    def list_revisions(self, experiment_uid, limit=None):
        """Print all revision for the given experiment uid in a table format.

        :param experiment_uid: Unique id of stored experiment
        :type experiment_uid: str

        :param limit: limit number of fetched records
        :type limit: int, optional

        **Example**

        .. code-block:: python

            client.experiments.list_revisions(experiment_uid)

        """
        ##For CP4D, check if either spce or project ID is set
        self._client._check_if_either_is_set()

        Experiments._validate_type(experiment_uid, u'experiment_uid', str, True)

        if not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES:
            raise WMLClientError(
                u'Revision support is not there in this WML server. It is supported only from 3.1.0 onwards.')
        else:
            url = self._client.service_instance._href_definitions.get_experiment_href(experiment_uid)
            experiment_resources = self._get_artifact_details(url, "revisions", limit, 'model revisions')[u'resources']
            experiment_values = [
                (m[u'metadata'][u'rev'], m[u'metadata'][u'name'], m[u'metadata'][u'created_at']) for m in
                experiment_resources]

            self._list(experiment_values, [u'GUID', u'NAME', u'CREATED'], limit, _DEFAULT_LIST_LENGTH)

    def clone(self, experiment_uid, space_id=None, action="copy", rev_id=None):
        """Create a new experiment identical with the given experiment either in the same space or in a new space.
        All dependent assets will be cloned too.

        :param experiment_uid:  Unique Id of the experiment to be cloned
        :type experiment_uid: str

        :param space_id: Unique Id of the space to which the experiment needs to be cloned
        :type space_id: str, optional

        :param action: action specifying "copy" or "move"
        :type action: str, optional

        :param rev_id: revision ID of the experiment
        :type rev_id: str, optional

        :return: metadata of the experiment cloned
        :rtype: dict

        **Example**

        .. code-block:: python

            client.experiments.clone(experiment_uid=artifact_id,space_id=space_uid,action="copy")

        .. note::

            * If revision id is not specified, all revisions of the artifact are cloned.

            * Space id is mandatory for move action.

        """
        Experiments._validate_type(experiment_uid, 'experiment_uid', str, True)
        clone_meta = {}
        if space_id is not None:
            clone_meta["space"] = {"href": API_VERSION + SPACES + "/" + space_id}
        if action is not None:
            clone_meta["action"] = action
        if rev_id is not None:
            clone_meta["rev"] = rev_id

        url = self._client.service_instance._href_definitions.get_experiment_href(experiment_uid)
        response_post = requests.post(url, json=clone_meta,
                                      headers=self._client._get_headers())

        details = self._handle_response(expected_status_code=200, operationName=u'cloning experiment',
                                        response=response_post)

        return details





#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2018- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

from ibm_watson_machine_learning.libs.repo.mlrepository import ModelArtifact


class HybridModelArtifact(ModelArtifact):
    """
    Class representing Hybrid model artifact
    """
    def __init__(self, uid, name, meta_props):
        """
        Constructor for Hybrid model artifact
        :param uid: unique id for Hybrid model artifact
        :param name: name of the model
        :param metaprops: properties of the model and model artifact
        """
        super(HybridModelArtifact, self).__init__(uid, name, meta_props)

